// Copyright 2022 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package loong64

// This order should be strictly consistent to that in a.out.go.
var cnames0 = []string{
	"NONE",
	"REG",
	"FREG",
	"FCSRREG",
	"FCCREG",
	"ZCON",
	"SCON",
	"UCON",
	"ADD0CON",
	"AND0CON",
	"ADDCON",
	"ANDCON",
	"LCON",
	"DCON",
	"SACON",
	"SECON",
	"LACON",
	"LECON",
	"DACON",
	"STCON",
	"SBRA",
	"LBRA",
	"SAUTO",
	"LAUTO",
	"SEXT",
	"LEXT",
	"ZOREG",
	"SOREG",
	"LOREG",
	"GOK",
	"ADDR",
	"TLS_LE",
	"TLS_IE",
	"GOTADDR",
	"TEXTSIZE",
	"NCLASS",
}
