// Copyright 2013 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package main

import "./a"

type S3 struct {
	p.S1
}

func main() {
	var i interface{} = S3{}
	_ = i
}
