// Copyright 2018 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// The runtime package uses //go:linkname to push the setEventHandler to this
// package.  To prevent the go tool from passing -complete to the compile tool,
// this file must remain stubbed out.
