// Copyright 2014 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package issue8828

//void foo();
import "C"

func Bar() {
	C.foo()
}
