// errorcheck

// Copyright 2016 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package p

type T interface { // ERROR "invalid recursive type: anonymous interface refers to itself"
	M(interface {
		T
	})
}
