// errorcheckdir

// Copyright 2011 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Test that unexported methods are not visible outside the package.
// Does not compile.

package ignored
