// rundir

// Copyright 2013 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Issue 6789: gccgo failed to find the hash function for an
// unexported struct embedded in an exported struct.

package ignored
