// Copyright 2023 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

#include "libcgo.h"

void x_cgo_getstackbound(uintptr bounds[2]) {} // no-op for now
