// rundir -ldflags -strictdups=2 -w=0

// Copyright 2019 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

//go:build !nacl && !js

package ignored
