// Copyright 2021 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package main

import "./a"

var m = a.I[int].M

var never bool

func main() {
	if never {
		m(nil)
	}
}
